;; BF LLVM Assembly Emitter
;; Jon Simons (simonsj at ccs dot neu dot edu)
;; November, 2008
;; -------------------------------------------------------------------------
;; This module emits LLVM Assembly for the BF IR.
;;
;;   A 30000 byte space is declared "internal constant" for use as
;;   the BF "memory".
;;
;;   Identifier %pIndex used as pointer to alloca'd space for the value
;;   of the absolute BF "pointer" value.
;;
;;   ANSI functions getchar and putchar are used for IN and OUT emitting.
;;   The declarations provided (with input, output types i8 instead of
;;   i32) are lying, but the LLVM tools seem to play nicely with this.
;;   Otherwise emitted code would need to sign extend and cast i8's to
;;   i32's.
;;
;;   This module emits code which is completely oblivious to any integer
;;   wrapping issues.
;;
;;   Emitted code was tested using llvm-as for assembling, and lli for
;;   running, from the LLVM 2.3 distribution.
;; -------------------------------------------------------------------------
(module bf-emit-llvm mzscheme
  
  (provide emit)
  
  ;; For AST structs
  (require "../bf-ast-to-ir.scm")
  ;; Emit Utils
  (require "bf-emit-util.scm")
  
  (require scheme/list)
  
  ;; emit : (list of EXPRs) -> string
  (define (emit AST)
    (string-append (emit-prologue)
                   (emit-program AST)
                   (emit-epilogue)))
  
  ;; emit-prologue : -> string
  (define (emit-prologue)
    (string-append "declare i32 @putchar(i8)~%"
                   "declare i8 @getchar()~%"
                   "@BFMem = internal constant [ 30000 x i8 ] zeroinitializer"
                   "~%"
                   "define i32 @main() {~%"
                   "  %pIndex = alloca i64~%"
                   "  store i64 0, i64* %pIndex~%"
                   "~%"))
  
  ;; Formatter string for the verbose getelementptr incantation.
  ;; get-ptr-str : string
  (define get-ptr-str
    "~a = getelementptr [ 30000 x i8 ]* @BFMem, i64 0, i64 ~a~%")
  
  ;; emit-epilogue : -> string
  (define (emit-epilogue)
    (string-append "  ret i32 0~%"
                   "}"))
  
  ;; emit-program : (list of EXPRs) -> string
  (define (emit-program AST)
    (apply string-append (emit-exprs AST #f 2 ())))
  
  ;; Emits a list of strings which correspond to the given
  ;; list of EXPRs, using the given indentation level.
  ;;
  ;; The argument "mult-mode" should be #f for normal emitting,
  ;; and a string representing the identifier for the multiplier
  ;; value otherwise.
  ;;
  ;; emit-exprs : (list of EXPRs) boolean number (list of strings)
  ;;              -> (list of strings)
  (define (emit-exprs lst mult-mode indent acc)
    ;; Here emit is a helper for padding and formatting
    ;; strings to be emitted
    (let ((emit  (case-lambda
                   (()   (error "emitting nothing?"))
                   ((x)  (pad x indent))
                   (args (pad (apply format args) indent)))))
      (if (null? lst)
          (reverse acc)
          (let* ((e   (car lst))
                 (ptr (EXPR-ptr e))
                 (N   (EXPR-N   e))
                 (str (case (EXPR-type e)
                        ((inc)
                         (let ((idx         (ssa-tmp))
                               (inc-idx     (ssa-tmp))
                               (inc-ptr     (ssa-tmp))
                               (curr-value  (ssa-tmp))
                               (mul-value   (ssa-tmp))
                               (new-value   (ssa-tmp)))
                           (string-append
                            (emit "~a = load i64* %pIndex~%" idx)
                            (emit "~a = add i64 ~a, ~v~%" inc-idx idx ptr)
                            (emit get-ptr-str inc-ptr inc-idx)
                            (emit "~a = load i8* ~a~%" curr-value inc-ptr)
                            (if mult-mode
                                (string-append
                                 (emit "~a = mul i8 ~a, ~v~%"
                                       mul-value
                                       mult-mode
                                       N)
                                 (emit "~a = add i8 ~a, ~a~%"
                                       new-value
                                       mul-value
                                       curr-value))
                                (emit "~a = add i8 ~a, ~v~%"
                                      new-value
                                      curr-value
                                      N))
                            (emit "store i8 ~a, i8* ~a~%" new-value inc-ptr))))
                        ((dec)
                         (let ((idx         (ssa-tmp))
                               (dec-idx     (ssa-tmp))
                               (dec-ptr     (ssa-tmp))
                               (curr-value  (ssa-tmp))
                               (mul-value   (ssa-tmp))
                               (new-value   (ssa-tmp)))
                           (string-append
                            (emit "~a = load i64* %pIndex~%" idx)
                            (emit "~a = add i64 ~a, ~v~%" dec-idx idx ptr)
                            (emit get-ptr-str dec-ptr dec-idx)
                            (emit "~a = load i8* ~a~%" curr-value dec-ptr)
                            (if mult-mode
                                (string-append
                                 (emit "~a = mul i8 ~a, ~v~%"
                                       mul-value
                                       mult-mode
                                       N)
                                 (emit "~a = sub i8 ~a, ~a~%"
                                       new-value
                                       curr-value
                                       mul-value))
                                (emit "~a = sub i8 ~a, ~v~%"
                                      new-value
                                      curr-value
                                      N))
                            (emit "store i8 ~a, i8* ~a~%" new-value dec-ptr))))
                        ((lt)   (let ((idx       (ssa-tmp))
                                      (old-value (ssa-tmp))
                                      (new-value (ssa-tmp)))
                                  (string-append
                                   (emit "~a = load i64* %pIndex~%" old-value)
                                   (emit "~a = sub i64 ~a, ~v~%"
                                         new-value
                                         old-value
                                         N)
                                   (emit "store i64 ~a, i64* %pIndex~%"
                                         new-value))))
                        ((rt)   (let ((idx       (ssa-tmp))
                                      (old-value (ssa-tmp))
                                      (new-value (ssa-tmp)))
                                  (string-append
                                   (emit "~a = load i64* %pIndex~%" old-value)
                                   (emit "~a = add i64 ~a, ~v~%"
                                         new-value
                                         old-value
                                         N)
                                   (emit "store i64 ~a, i64* %pIndex~%"
                                         new-value))))
                        ((out)
                         (let ((idx      (ssa-tmp))
                               (out-idx  (ssa-tmp))
                               (idx-ptr  (ssa-tmp))
                               (curr-val (ssa-tmp)))
                           (string-append
                            (emit "~a = load i64* %pIndex~%" idx)
                            (emit "~a = add i64 ~a, ~v~%" out-idx idx ptr)
                            (emit get-ptr-str idx-ptr out-idx)
                            (emit "~a = load i8* ~a~%" curr-val idx-ptr)
                            (emit "call i32 @putchar(i8 ~a)~%" curr-val))))
                        ((in)   
                         (let ((idx      (ssa-tmp))
                               (in-idx   (ssa-tmp))
                               (idx-ptr  (ssa-tmp))
                               (new-val  (ssa-tmp)))
                           (string-append
                            (emit "~a = load i64* %pIndex~%" idx)
                            (emit "~a = add i64 ~a, ~v~%" in-idx idx ptr)
                            (emit get-ptr-str idx-ptr in-idx)
                            (emit "~a = call i8 @getchar()~%" new-val)
                            (emit "store i8 ~a, i8* ~a~%" new-val idx-ptr))))
                        ((loop)  (let* ((loop-tmp (gen-loop))
                                        (start    (GENLOOP-start loop-tmp))
                                        (do       (GENLOOP-do    loop-tmp))
                                        (end      (GENLOOP-end   loop-tmp))
                                        (idx      (ssa-tmp))
                                        (curr-idx (ssa-tmp))
                                        (curr-ptr (ssa-tmp))
                                        (curr-val (ssa-tmp))
                                        (cmp-tmp  (ssa-tmp)))
                                   (string-append
                                    (emit "br label %~a~%" start)
                                    (format "~a:~%" start)
                                    (emit "~a = load i64* %pIndex~%" idx)
                                    (emit "~a = add i64 ~a, ~v~%"
                                          curr-idx
                                          idx
                                          ptr)
                                    (emit get-ptr-str curr-ptr curr-idx)
                                    (emit "~a = load i8* ~a~%"
                                          curr-val
                                          curr-ptr)
                                    (emit "~a = icmp eq i8 0, ~a~%"
                                          cmp-tmp
                                          curr-val)
                                    (emit "br i1 ~a, label %~a, label %~a~%"
                                          cmp-tmp
                                          end
                                          do)
                                    (format "~a:~%" do)
                                    (apply string-append
                                           (emit-exprs (EXPR-exprs e)
                                                       #f
                                                       indent
                                                       ()))
                                    (emit "br label %~a~%" start)
                                    (format "~a:~%" end))))
                        ((mul)  
                         (let ((skip-label (mul-skip))
                               (do-label   (do-mul-skip))
                               (idx        (ssa-tmp))
                               (mul-idx    (ssa-tmp))
                               (mul-ptr    (ssa-tmp))
                               (mul-val    (ssa-tmp))
                               (cmp-tmp    (ssa-tmp)))
                           (string-append
                            (emit "~a = load i64* %pIndex~%" idx)
                            (emit "~a = add i64 ~a, ~v~%" mul-idx idx ptr)
                            (emit get-ptr-str mul-ptr mul-idx)
                            (emit "~a = load i8* ~a~%" mul-val mul-ptr)
                            (emit "~a = icmp eq i8 0, ~a~%" cmp-tmp mul-val)
                            (emit "br i1 ~a, label %~a, label %~a~%"
                                  cmp-tmp
                                  skip-label
                                  do-label)
                            (format "~a:~%" do-label)
                            (apply string-append
                                   (emit-exprs (EXPR-exprs e)
                                               ;; Note mult-mode overload
                                               mul-val
                                               indent
                                               ()))
                            (emit "br label %~a~%" skip-label)
                            (format "~a:~%" skip-label))))
                        ((zero)
                         (let ((idx      (ssa-tmp))
                               (zero-idx (ssa-tmp))
                               (zero-ptr (ssa-tmp)))
                           (string-append
                            (emit "~a = load i64* %pIndex~%" idx)
                            (emit "~a = add i64 ~a, ~v~%" zero-idx idx ptr)
                            (emit get-ptr-str zero-ptr zero-idx)
                            (emit "store i8 0, i8* ~a~%" zero-ptr))))
                        (else  (error "unknown token, can't emit")))))
            (emit-exprs (cdr lst) mult-mode indent (cons str acc))))))
  
  ;; GENSSATMP Stuff ---------------------------------------------------------
  (define (gen-ssa-tmp)
    (let ((tmp-num (box 0)))
      (lambda ()
        (begin
          (set-box! tmp-num (+ 1 (unbox tmp-num)))
          (format "%tmp~v" (unbox tmp-num))))))
  
  ;; Used for generating SSA-friendly identifiers
  (define ssa-tmp (gen-ssa-tmp))
  
  )
